;=============================================================================
;          ProTracker 2.3D i386 asm replayer port, by 8bitbubsy 2022
;
; NOTE:
;  Some stuff has been modified for speed, but the result is the same.
;
; TABSIZE = 8 spaces
;=============================================================================

.386p
.MODEL FLAT

	INCLUDE "apaula.i"
	INCLUDE "areplay.i"

	; replayer.c
	EXTERN _Replayer_Update:PROC
	EXTERN _Replayer_SetBPM:PROC

	; apaula.asm
	EXTERN _Paula_SetDMACON:PROC
	EXTERN _Paula_SetDat:PROC
	EXTERN _Paula_SetLen:PROC
	EXTERN _Paula_SetPer:PROC
	EXTERN _Paula_SetVol:PROC

	; audio.c
	EXTERN _Audio_ClearSampleCounter:PROC

	; mouse.c
	EXTERN _Mouse_UpdateColor:PROC

	PUBLIC _Replayer_ConvertPatternData
	PUBLIC _Replayer_PrepareSetPos
	PUBLIC _Replayer_PreparePlay
	PUBLIC _Replayer_PrepareStop
	PUBLIC _Replayer_StepPlayForwards
	PUBLIC _Replayer_StepPlayBackwards
	PUBLIC _Replayer_Tick

.CODE

;=============================================================================
;                                  FUNCTIONS
;=============================================================================

	; Input:
	;  ESI = Pointer to .MOD pattern data
	;  EAX = Number of patterns
	ALIGN 4
_Replayer_ConvertPatternData PROC NEAR
	pushad
	; --------------------
	shl eax,10
	lea edi,[esi+eax]	; edi = end of pattern data
	; --------------------
	mov ecx,36		; notes to scan (period -> note)
	xor edx,edx		; hi-word of edx must be zeroed
	mov ebp,0FFFh		; period AND-mask
	; --------------------
ploop1:	mov eax,[esi]
	mov ebx,eax
	shr ebx,16
	; --------------------
	; period
	mov dl,ah
	mov dh,al
	and edx,ebp		; ebp = 0FFFh
	; --------------------
	; sample
	and al,0F0h
	mov ah,bl
	shr ah,4
	or al,ah
	and al,31
	mov [esi+1],al
	; --------------------
	; effect
	and bl,15
	mov [esi+2],bl
	; --------------------
	; effect parameter
	mov [esi+3],bh
	; --------------------
	; convert period to note
	xor eax,eax
	test edx,edx
	jz noNote
	lea ebx,PeriodTable
ploop2:	cmp dx,[ebx+eax*2]
	jae pfound
	inc eax
	cmp eax,ecx		; ecx = 36
	jne ploop2
pfound:	inc eax			; 1..36, 37 = illegal
	; --------------------
noNote:	mov [esi+0],al
	; --------------------
	add esi,4
	cmp esi,edi
	jne ploop1
	; --------------------
	popad
	retn
_Replayer_ConvertPatternData ENDP

	; Input:
	;  EAX = Order
	ALIGN 4
_Replayer_PrepareSetPos PROC NEAR
	push eax
	push ebp
	mov ebp,_moduleOffset
	; -----------------------------
	and eax,7Fh
	mov SongPosition,al
	mov m_order[ebp],al
	movzx eax,byte ptr m_orderlist[ebp+eax]
	mov m_pattern[ebp],al
	shl eax,10
	mov ebp,_patternDataOffset
	add ebp,eax
	mov PatternPtr,ebp
	; -----------------------------
	xor eax,eax
	mov PBreakPosition,al
	mov PBreakFlag,al
	mov PosJumpAssert,al
	mov PattDelayTime,al
	mov PattDelayTime2,al
	; -----------------------------
	pop ebp
	pop eax
	retn
_Replayer_PrepareSetPos ENDP

	ALIGN 4
_Replayer_PreparePlay PROC NEAR
	push eax
	push ebp
	mov ebp,_moduleOffset
	; -----------------------------
	mov al,m_order[ebp]
	mov SongPosition,al
	mov al,m_row[ebp]
	mov PatternPosition,al
	; -----------------------------
	xor eax,eax
	mov PBreakPosition,al
	mov PosJumpAssert,al
	mov PBreakFlag,al
	mov SetBPMFlag,al
	; -----------------------------
	movzx eax,byte ptr m_pattern[ebp]
	shl eax,10
	mov ebp,_patternDataOffset
	add ebp,eax
	mov PatternPtr,ebp
	; -----------------------------
	pop ebp
	pop eax
	retn
_Replayer_PreparePlay ENDP

	ALIGN 4
_Replayer_StepPlayForwards PROC NEAR
	push eax
	push ebp
	mov ebp,_moduleOffset
	; -----------------------------
	mov byte ptr m_tick[ebp],0
	mov byte ptr StepPlayEnable,1
	; -----------------------------
	mov al,m_order[ebp]
	mov SongPosition,al
	mov al,m_row[ebp]
	mov PatternPosition,al
	; -----------------------------
	xor eax,eax
	mov PBreakPosition,al
	mov PosJumpAssert,al
	mov PBreakFlag,al
	mov SetBPMFlag,al
	; -----------------------------
	movzx eax,byte ptr m_pattern[ebp]
	shl eax,10
	mov ebp,_patternDataOffset
	add ebp,eax
	mov PatternPtr,ebp
	; -----------------------------
	call _Audio_ClearSampleCounter
	mov _Replayer_PlayMode,PLAY_MODE_PATTERN
	mov _Replayer_SongPlaying,1
	; -----------------------------
SPfLoop:
	cmp byte ptr StepPlayEnable,1
	je SPfLoop
	; -----------------------------
	pop ebp
	pop eax
	retn
_Replayer_StepPlayForwards ENDP

	ALIGN 4
_Replayer_StepPlayBackwards PROC NEAR
	push eax
	push ebp
	mov ebp,_moduleOffset
	; -----------------------------
	mov byte ptr m_tick[ebp],0
	mov byte ptr StepPlayEnable,1
	; -----------------------------
	mov al,m_order[ebp]
	mov SongPosition,al
	mov al,m_row[ebp]
	mov PatternPosition,al
	; -----------------------------
	xor eax,eax
	mov PBreakPosition,al
	mov PosJumpAssert,al
	mov PBreakFlag,al
	mov SetBPMFlag,al
	; -----------------------------
	movzx eax,byte ptr m_pattern[ebp]
	shl eax,10
	mov ebp,_patternDataOffset
	add ebp,eax
	mov PatternPtr,ebp
	; -----------------------------
	call _Audio_ClearSampleCounter
	mov _Replayer_PlayMode,PLAY_MODE_PATTERN
	mov _Replayer_SongPlaying,1
	; -----------------------------
SPBLoop:
	cmp byte ptr StepPlayEnable,1
	je SPBLoop
	; -----------------------------
	mov ebp,_moduleOffset
	mov al,m_row[ebp]
	sub al,2
	and al,63
	mov m_row[ebp],al
	; -----------------------------
	pop ebp
	pop eax
	retn
_Replayer_StepPlayBackwards ENDP

	ALIGN 4
_Replayer_PrepareStop PROC NEAR
	push eax
	push ebp
	mov ebp,_moduleOffset
	mov al,m_order[ebp]
	mov SongPosition,al
	mov al,m_row[ebp]
	mov PatternPosition,al
	xor eax,eax
	mov PattDelayTime,al
	mov PattDelayTime2,al
	pop ebp
	pop eax
	jmp RestoreEffects2
_Replayer_PrepareStop ENDP

	ALIGN 4
_Replayer_Tick PROC NEAR
	pushad
	xor edx,edx	; dl = 0
	; -----------------------------------------------
	; Handle BPM updates from last tick here, since
	; CIA timer chip writes are updated after the
	; current cycle is done. I.e. one tick delayed.
	; -----------------------------------------------
	cmp SetBPMFlag,dl
	je NoBPMUpdate
	movzx eax,byte ptr SetBPMFlag
	push eax
	mov SetBPMFlag,dl
	call _Replayer_SetBPM
	add esp,1*4
NoBPMUpdate:
	mov ebp,_moduleOffset
	; -----------------------------------------------
	cmp _Replayer_PlayMode,PLAY_MODE_IDLE
	je NoNewPositionYet
	; -----------------------------------------------
	; Increase playback timer if in "play song" mode
	; -----------------------------------------------
	cmp _Replayer_PlayMode,PLAY_MODE_SONG
	jne NoTimeInc
	; -----------------------------------------------
	cmp byte ptr m_vblanktempo[ebp],1
	jne ciaTempo
	mov ecx,256-32	; use vblank BPM entry (~49.92Hz)
	jmp GetDuration
ciaTempo:
	; -----------------------------------------------
	movzx ecx,byte ptr m_bpm[ebp]
	cmp cl,32
	jb NoTimeInc ; error! BPM < 32 (shouldn't happen)
	sub ecx,32
GetDuration:
	mov eax,[MusicTimeTab32+ecx*4]
	add _Replayer_PlaybackSecondsFrac,eax
	adc _Replayer_PlaybackSeconds,0
	; -----------------------------------------------
NoTimeInc:
	xor edx,edx ; dl = 0
	cmp StepPlayEnable,dl
	jne StepPlay
	inc byte ptr m_tick[ebp]
	mov al,m_tick[ebp]
	cmp al,m_speed[ebp]
	jb NoNewNote
StepPlay:
	mov m_tick[ebp],dl
	cmp PattDelayTime2,dl
	je GetNewNote
	call NoNewAllChannels
	jmp dskip

NoNewNote:
	call NoNewAllChannels
	jmp NoNewPositionYet

NoNewAllChannels:
	mov edi,_channelOffset
	mov esi,_paulaOffset
	; ----------------------------------
	; channel #1
	call CheckEffects
	add edi,n_sizeof
	add esi,v_sizeof

	; channel #2
	call CheckEffects
	add edi,n_sizeof
	add esi,v_sizeof

	; channel #3
	call CheckEffects
	add edi,n_sizeof
	add esi,v_sizeof

	; channel #4
	jmp CheckEffects
	; ----------------------------------

GetNewNote:
	; ----------------------------------
	; update tracker pattern display
	push byte ptr SongPosition
	push byte ptr PatternPosition
	call _Replayer_Update
	add esp,2*4
	; ----------------------------------
	movzx eax,byte ptr PatternPosition
	shl eax,4
	mov ebx,PatternPtr
	add ebx,eax
	mov PattReadPtr,ebx
	mov edi,_channelOffset
	mov esi,_paulaOffset
	mov word ptr DMACONtemp,0
	; ----------------------------------
	; channel #1
	call PlayVoice
	movzx eax,byte ptr n_volume[edi]
	call _Paula_SetVol
	add edi,n_sizeof
	add esi,v_sizeof

	; channel #2
	call PlayVoice
	movzx eax,byte ptr n_volume[edi]
	call _Paula_SetVol
	add edi,n_sizeof
	add esi,v_sizeof

	; channel #3
	call PlayVoice
	movzx eax,byte ptr n_volume[edi]
	call _Paula_SetVol
	add edi,n_sizeof
	add esi,v_sizeof

	; channel #4
	call PlayVoice
	movzx eax,byte ptr n_volume[edi]
	call _Paula_SetVol
	; ----------------------------------
	jmp SetDMA

PlayVoice:
	mov eax,[edi]
	and eax,0FFFF00FFh ; test n_note, n_cmd and n_param (not n_samplenum)
	jnz plvskip
	call PerNop
plvskip:
	;call CheckMetronome
	mov eax,PattReadPtr
	mov eax,[eax] ; Read note from pattern
	add dword ptr PattReadPtr,4
	mov [edi],eax
	; ---------------------
	movzx eax,byte ptr n_samplenum[edi]
	test eax,eax
	jz SetRegisters		; Instrument was zero
	dec eax
	mov ecx,_sampleOffsets
	mov ecx,[ecx+eax*4]	; ecx points to sample struct
	; ---------------------
	mov eax,s_data[ecx]
	mov n_start[edi],eax
	mov ax,s_length[ecx]
	mov n_length[edi],ax
	mov al,s_finetune[ecx]
	mov n_finetune[edi],al
	mov al,s_volume[ecx]
	mov n_volume[edi],al
	mov dx,word ptr s_replen[ecx]
	mov n_replen[edi],dx
	movzx eax,word ptr s_repeat[ecx]
	test eax,eax
	jz NoLoop
	mov ebx,n_start[edi]
	add ebx,eax		; Add repeat
	add ebx,eax		; Add repeat (again)
	mov n_loopstart[edi],ebx
	mov n_wavestart[edi],ebx
	add ax,dx		; Add replen
	mov n_length[edi],ax
	jmp SetRegisters

NoLoop:
	mov ebx,n_start[edi]
	mov n_loopstart[edi],ebx
	mov n_wavestart[edi],ebx

SetRegisters:
	cmp byte ptr n_note[edi],0
	je CheckMoreEffects	; If no note

	mov ax,n_cmd[edi]
	and ax,0F00Fh
	cmp ax,0500Eh		; finetune (E5x)
	je DoSetFineTune
	cmp al,3		; TonePortamento
	je ChkTonePorta
	cmp al,5		; TonePortamento + VolSlide
	je ChkTonePorta
	cmp al,9		; Sample Offset
	jne SetPeriod
	call CheckMoreEffects
	jmp SetPeriod

DoSetFineTune:
	call SetFineTune
	jmp SetPeriod

ChkTonePorta:
	call SetTonePorta
	jmp CheckMoreEffects

SetPeriod:
	movzx eax,byte ptr n_finetune[edi]
	mov eax,[TunedPeriodPtr+eax*4]
	movzx ebx,byte ptr n_note[edi]	; always >0 at this point
	dec ebx
	mov ax,[eax+ebx*2]
	mov n_period[edi],ax

	mov ax,n_cmd[edi]
	and ax,0F00Fh
	cmp ax,0D00Eh ; Note Delay (EDx)?
	je CheckMoreEffects

	; stop Paula DMA
	movzx eax,word ptr n_dmabit[edi]
	call _Paula_SetDMACON

	test byte ptr n_wavecontrol[edi],4
	jnz vibnoc
	mov byte ptr n_vibratopos[edi],0
vibnoc:	test byte ptr n_wavecontrol[edi],64
	jnz trenoc
	mov byte ptr n_tremolopos[edi],0
trenoc:
	; set Paula length
	movzx eax,word ptr n_length[edi]
	call _Paula_SetLen

	; set Paula data
	mov eax,n_start[edi]
	call _Paula_SetDat
	
	mov eax,n_start[edi]
	test eax,eax
	jnz sdmaskp
	xor eax,eax
	mov n_loopstart[edi],eax
	inc eax	; eax = 1
	call _Paula_SetLen
	mov n_replen[edi],ax
sdmaskp:

	; set Paula period
	movzx eax,word ptr n_period[edi]
	call _Paula_SetPer
	
	; prepare DMACONtemp for SetDMA later
	mov ax,n_dmabit[edi]
	or DMACONtemp,ax

	jmp CheckMoreEffects

SetDMA:
	; start selected Paula DMAs
	movzx eax,word ptr DMACONtemp
	;and ax,ActiveChannels
	or ax,8000h ; set
	call _Paula_SetDMACON
	
	; now set new data/length for after first Paula cycle

	mov edi,_channelOffset
	mov esi,_paulaOffset

	; channel #1
	mov eax,n_loopstart[edi]
	call _Paula_SetDat
	movzx eax,word ptr n_replen[edi]
	call _Paula_SetLen
	add edi,n_sizeof
	add esi,v_sizeof

	; channel #2
	mov eax,n_loopstart[edi]
	call _Paula_SetDat
	movzx eax,word ptr n_replen[edi]
	call _Paula_SetLen
	add edi,n_sizeof
	add esi,v_sizeof

	; channel #3
	mov eax,n_loopstart[edi]
	call _Paula_SetDat
	movzx eax,word ptr n_replen[edi]
	call _Paula_SetLen
	add edi,n_sizeof
	add esi,v_sizeof

	; channel #4
	mov eax,n_loopstart[edi]
	call _Paula_SetDat
	movzx eax,word ptr n_replen[edi]
	call _Paula_SetLen

dskip:	xor edx,edx	; dl = 0
	inc byte ptr PatternPosition
	cmp PattDelayTime,dl
	je dskpc
	mov al,PattDelayTime
	mov PattDelayTime2,al
	mov PattDelayTime,dl
dskpc:	cmp PattDelayTime2,dl
	je dskpa
	dec byte ptr PattDelayTime2
	jz dskpa
	dec byte ptr PatternPosition
dskpa:	cmp PBreakFlag,dl
	je nnpysk
	mov PBreakFlag,dl
	mov al,PBreakPosition
	mov PatternPosition,al
	mov PBreakPosition,dl
nnpysk:	cmp StepPlayEnable,dl
	je nnpysk2
	call DoStopIt
	mov StepPlayEnable,dl
	and byte ptr PatternPosition,63
	push byte ptr SongPosition
	push byte ptr PatternPosition
	call _Replayer_Update
	add esp,2*4
nnpysk2:
	cmp byte ptr PatternPosition,64
	jne NoNewPositionYet
NextPosition:
	mov al,PBreakPosition
	mov PatternPosition,al
	mov PBreakPosition,dl
	mov PosJumpAssert,dl

	; if in "play pattern" or "record" mode, don't increase song position
	mov al,_Replayer_PlayMode
	cmp al,PLAY_MODE_PATTERN
	je NoNewPositionYet
	cmp al,PLAY_MODE_RECORD
	je NoNewPositionYet

	movzx eax,byte ptr SongPosition
	inc eax
	and eax,127
	cmp al,m_length[ebp]
	jb npskip
	xor eax,eax
npskip:	mov SongPosition,al
	movzx eax,byte ptr m_orderlist[ebp+eax]
	shl eax,10
	mov edi,_patternDataOffset
	add edi,eax
	mov PatternPtr,edi

	cmp StepPlayEnable,dl
	je NoNewPositionYet
	call DoStopIt
	mov StepPlayEnable,dl
	and byte ptr PatternPosition,63
	push byte ptr SongPosition
	push byte ptr PatternPosition
	call _Replayer_Update
	add esp,2*4

NoNewPositionYet:
	cmp byte ptr PosJumpAssert,0
	jne NextPosition
	popad
	retn

DoStopIt:
	mov byte ptr _Replayer_PlayMode,PLAY_MODE_IDLE
	mov byte ptr _Replayer_SongPlaying,0
	push esi
	push edi
	push ebp
	call _Mouse_UpdateColor	; interrupt safe
	pop ebp
	pop edi
	pop esi
	xor edx,edx	; dl = 0
	mov PattDelayTime,dl
	mov PattDelayTime2,dl
	jmp RestoreEffects2

RestoreEffects2:
	push ecx
	push edx
	push edi
	mov edi,_channelOffset
	xor edx,edx
	; --------------------------
	mov ecx,4
re2l:	mov n_wavecontrol[edi],dl
	mov n_glissfunk[edi],dl
	mov n_finetune[edi],dl
	mov n_loopcount[edi],dl
	add edi,n_sizeof
	dec ecx
	jnz re2l
	; --------------------------
	pop edi
	pop edx
	pop ecx
	retn

; ----------------------------------------------------------
; ----------------------------------------------------------

CheckEffects:
	mov eax,n_chanindex[edi]
	cmp byte ptr m_muted[ebp+eax],1
	je Return ; channel muted?
	call chkefx2
	movzx eax,n_volume[edi]
	jmp _Paula_SetVol

chkefx2:
	call UpdateFunk	
	cmp word ptr n_cmd[edi],0	; n_cmd & n_param = 0?
	je Return
	movzx ebx,byte ptr n_cmd[edi]
	movzx eax,byte ptr n_param[edi]
	jmp [JumpList1+ebx*4]

SetBack:
	movzx eax,word ptr n_period[edi]
	call _Paula_SetPer
	; -----------------------------
	movzx eax,byte ptr n_param[edi]
	mov bl,n_cmd[edi]
	cmp bl,07h
	je Tremolo
	cmp bl,0Ah
	je VolumeSlide
Return:	retn

PerNop:
	movzx eax,word ptr n_period[edi]
	jmp _Paula_SetPer

Arpeggio:
	mov ecx,eax	; ecx = effect parameter (note nybbles)
	movzx ax,byte ptr m_tick[ebp]
	mov bl,3
	div bl		; only 14 cycles on a 386
	cmp ah,1
	je Arpeggio1
	cmp ah,2
	je Arpeggio2
Arpeggio0:
	movzx eax,word ptr n_period[edi]
	jmp _Paula_SetPer
Arpeggio1:
	shr ecx,4
	jmp ArpeggioFind
Arpeggio2:
	and ecx,15
ArpeggioFind:
	movzx ebx,byte ptr n_finetune[edi]
	mov ebx,[TunedPeriodPtr+ebx*4]
	mov eax,37
afloop:	mov dx,[ebx]
	cmp dx,n_period[edi]
	jbe ArpeggioSet
	add ebx,2
	dec eax
	jnz afloop
	retn
ArpeggioSet:
	movzx eax,word ptr [ebx+ecx*2]
	jmp _Paula_SetPer
	
FinePortaUp:
	cmp byte ptr m_tick[ebp],0
	jne Return
PortaUp:
	sub n_period[edi],ax
	mov bx,n_period[edi]
	and bx,0FFFh
	cmp bx,113
	jns PortaUskip	; this can and will underflow =)
	and word ptr n_period[edi],0F000h
	or word ptr n_period[edi],113
PortaUskip:
	movzx eax,word ptr n_period[edi]
	and eax,0FFFh
	jmp _Paula_SetPer

FinePortaDown:
	cmp byte ptr m_tick[ebp],0
	jne Return
PortaDown:
	add n_period[edi],ax
	mov bx,n_period[edi]
	and bx,0FFFh
	cmp bx,856
	js PortaUskip
	and word ptr n_period[edi],0F000h
	or word ptr n_period[edi],856
PortaDskip:
	movzx eax,word ptr n_period[edi]
	and eax,0FFFh
	jmp _Paula_SetPer

SetTonePorta:
	xor ecx,ecx
	movzx eax,byte ptr n_note[edi]
	or eax,eax
	je PortaSkip
	movzx ebx,byte ptr n_finetune[edi]
	mov ebx,[TunedPeriodPtr+ebx*4]
	; ---------------------------------
	cmp al,37
	je PortaIllegal
	dec eax
	mov cx,[ebx+eax*2]
	jmp PortaSkip
PortaIllegal:
	mov cx,[ebx+(35*2)] ; note B
PortaSkip:
	mov n_wantedperiod[edi],cx
	mov byte ptr n_toneportdirec[edi],0
	cmp cx,n_period[edi]
	je ClearTonePorta
	jge Return
	mov byte ptr n_toneportdirec[edi],1
	retn
ClearTonePorta:
	mov word ptr n_wantedperiod[edi],0
	retn

TonePortamento:
	test al,al
	jz TonePortNoChange
	mov n_toneportspeed[edi],al
	mov byte ptr n_param[edi],0
TonePortNoChange:
	cmp word ptr n_wantedperiod[edi],0
	je Return
	movzx ax,byte ptr n_toneportspeed[edi]
	mov bx,n_period[edi]
	cmp byte ptr n_toneportdirec[edi],0
	jnz TonePortaUp
TonePortaDown:
	add bx,ax
	cmp bx,n_wantedperiod[edi]
	jl TonePortaSetPer
TonePortaClear:
	mov bx,n_wantedperiod[edi]
	mov word ptr n_wantedperiod[edi],0
	jmp TonePortaSetPer
TonePortaUp:
	sub bx,ax
	cmp bx,n_wantedperiod[edi]
	jl TonePortaClear
TonePortaSetPer:
	mov n_period[edi],bx
	test byte ptr n_glissfunk[edi],0Fh
	jz GlissSkip
	movzx edx,byte ptr n_finetune[edi]
	mov edx,[TunedPeriodPtr+edx*4]
	xor ecx,ecx
GlissLoop:
	cmp bx,[edx+ecx]
	jae GlissFound
	add ecx,2
	cmp ecx,37*2
	jb GlissLoop
	mov cx,35*2
GlissFound:
	mov bx,[edx+ecx]
GlissSkip:
	movzx eax,bx
	jmp _Paula_SetPer

Vibrato:
	test al,al
	jz Vibrato2
	mov bl,n_vibratocmd[edi]
	and al,00Fh
	jz vibskip
	and bl,0F0h
	or bl,al
vibskip:
	mov al,n_param[edi]
	and al,0F0h
	jz vibskip2
	and bl,00Fh
	or bl,al
vibskip2:
	mov n_vibratocmd[edi],bl
Vibrato2:
	movzx eax,byte ptr n_vibratopos[edi]
	shr al,2
	and al,1Fh
	mov bl,n_wavecontrol[edi]
	and bl,3
	jz vib_sine
	shl al,3
	cmp bl,1
	je vib_rampdown
	mov bl,255
	jmp vib_set
vib_rampdown:
	cmp byte ptr n_vibratopos[edi],0
	jns vib_rampdown2
	mov bl,255
	sub bl,al
	jmp vib_set
vib_rampdown2:
	mov bl,al
	jmp vib_set
vib_sine:
	mov bl,[VibratoTable+eax]
vib_set:
	mov al,n_vibratocmd[edi]
	and al,15
	mul bl
	shr ax,7
	mov bx,n_period[edi]
	cmp byte ptr n_vibratopos[edi],0
	js VibratoNeg
	add bx,ax
	jmp Vibrato3
VibratoNeg:
	sub bx,ax
Vibrato3:
	mov cl,n_vibratocmd[edi]
	shr cl,2
	and cl,3Ch
	add byte ptr n_vibratopos[edi],cl
	movzx eax,bx
	jmp _Paula_SetPer

TonePlusVolSlide:
	call TonePortNoChange
	mov al,n_param[edi]
	jmp VolumeSlide

VibratoPlusVolSlide:
	call Vibrato2
	mov al,n_param[edi]
	jmp VolumeSlide

Tremolo:
	test al,al
	jz Tremolo2
	mov bl,n_tremolocmd[edi]
	and al,00Fh
	jz treskip
	and bl,0F0h
	or bl,al
treskip:
	mov al,n_param[edi]
	and al,0F0h
	jz treskip2
	and bl,00Fh
	or bl,al
treskip2:
	mov n_tremolocmd[edi],bl
Tremolo2:
	movzx eax,byte ptr n_tremolopos[edi]
	shr al,2
	and al,1Fh
	mov bl,n_wavecontrol[edi]
	shr bl,4
	and bl,3
	jz tre_sine
	shl al,3
	cmp bl,1
	je tre_rampdown
	mov bl,255
	jmp tre_set
tre_rampdown:
	cmp byte ptr n_vibratopos[edi],0 ; yes, an error from the PT code!
	jns tre_rampdown2
	mov bl,255
	sub bl,al
	jmp tre_set
tre_rampdown2:
	mov bl,al
	jmp tre_set
tre_sine:
	mov bl,[VibratoTable+eax]
tre_set:
	mov al,n_tremolocmd[edi]
	and al,15
	mul bl
	shr ax,6
	movzx bx,byte ptr n_volume[edi]
	cmp byte ptr n_tremolopos[edi],0
	js TremoloNeg
	add bx,ax
	cmp bx,64
	jbe TremoloOk
	mov bx,64
	jmp TremoloOk
TremoloNeg:
	sub bx,ax
	jns TremoloOk
	xor ebx,ebx
TremoloOk:
	mov cl,n_tremolocmd[edi]
	shr cl,2
	and cl,3Ch
	add byte ptr n_tremolopos[edi],cl
	movzx eax,bx
	call _Paula_SetVol
	add esp,4 ; don't go back to CheckEffects (prevent volume override)
	retn

SampleOffset:
	test al,al
	jz sononew
	mov n_sampleoffset[edi],al
sononew:
	mov al,n_sampleoffset[edi]
	shl ax,7
	cmp ax,n_length[edi]
	jge sofskip	; PT bug: signed comparison instead of unsigned!
	sub n_length[edi],ax
	add ax,ax
	add n_start[edi],eax
	retn
sofskip:
	mov word ptr n_length[edi],1
	retn

VolumeSlide:
	shr al,4
	jz VolSlideDown
VolSlideUp:
	mov bl,n_volume[edi]
	add bl,al
	cmp bl,64
	js vsuskip
	mov bl,64
vsuskip:
	mov n_volume[edi],bl
	retn

VolSlideDown:
	mov al,n_param[edi]
	and al,15
VolSlideDown2:
	mov bl,n_volume[edi]
	sub bl,al
	jns vsdskip
	xor ebx,ebx
vsdskip:
	mov n_volume[edi],bl
	retn

PositionJump:
	dec al
	mov SongPosition,al
pj2:	mov byte ptr PBreakPosition,0
	mov byte ptr PosJumpAssert,1
	retn

VolumeChange:
	cmp al,64
	jb VolumeOk
	mov al,64
VolumeOk:
	mov n_volume[edi],al
	retn

PatternBreak:
	mov bl,al
	shr al,4
	mov cl,10
	mul cl
	and bl,15
	add al,bl
	cmp al,63
	ja pj2 ; >63, jump to 00 instead
	mov PBreakPosition,al
	mov byte ptr PosJumpAssert,1
	retn

SetSpeed:
	test al,al
	jz SpeedNull
	cmp byte ptr m_vblanktempo[ebp],1
	je normspd
	cmp al,32
	jb normspd
	mov byte ptr m_tick[ebp],0
	mov SetBPMFlag,al
	retn
normspd:
	mov byte ptr m_speed[ebp],al
	retn
SpeedNull:
	mov byte ptr _Replayer_PlayMode,PLAY_MODE_IDLE
	mov byte ptr _Replayer_SongPlaying,0
	push esi
	push edi
	push ebp
	call _Mouse_UpdateColor	; interrupt safe
	pop ebp
	pop edi
	pop esi
	retn

CheckMoreEffects:
	movzx ebx,byte ptr n_cmd[edi]
	movzx eax,byte ptr n_param[edi]
	jmp [JumpList2+ebx*4]
	
E_Commands:
	mov ebx,eax
	shr ebx,4	; command
	and eax,15	; param
	jmp [E_JumpList+ebx*4]
	
FilterOnOff:
	retn

SetGlissControl:
	and byte ptr n_glissfunk[edi],0F0h
	or n_glissfunk[edi],al
	retn

SetVibratoControl:
	and byte ptr n_wavecontrol[edi],0F0h
	or n_wavecontrol[edi],al
	retn

SetFineTune:
	mov byte ptr n_finetune[edi],al
	retn

JumpLoop:
	cmp byte ptr m_tick[ebp],0
	jne Return
	test al,al
	jz SetLoop
	mov bl,n_loopcount[edi]
	test bl,bl
	jz jumpcnt
	dec byte ptr n_loopcount[edi]
	jz Return
jmploop:
	mov al,n_pattpos[edi]
	mov PBreakPosition,al
	mov byte ptr PBreakFlag,1
	retn
jumpcnt:
	mov n_loopcount[edi],al
	jmp jmploop
SetLoop:
	mov al,PatternPosition
	and al,63
	mov n_pattpos[edi],al
	retn

SetTremoloControl:
	shl al,4
	and byte ptr n_wavecontrol[edi],15
	or n_wavecontrol[edi],al
	retn

KarplusStrong:
	; I didn't implement this... Worst and least used effect ever!
	retn

RetrigNote:
	test al,al
	jz Return
	mov bl,al	; bl = retrig tick (effect parameter)
	movzx ax,byte ptr m_tick[ebp]
	test al,al
	jne rtnskp
	cmp byte ptr n_note[edi],0
	jne Return
rtnskp:	div bl		; only 14 cycles on a 386
	test ah,ah
	jnz Return
DoRetrg:
	; stop Paula DMA
	movzx eax,word ptr n_dmabit[edi]
	call _Paula_SetDMACON

	; set Paula data
	mov eax,n_start[edi]
	call _Paula_SetDat

	; set Paula length
	movzx eax,word ptr n_length[edi]
	call _Paula_SetLen

	; set Paula period
	movzx eax,word ptr n_period[edi]
	call _Paula_SetPer

	; start Paula DMA
	movzx eax,word ptr n_dmabit[edi]
	or ax,8000h
	call _Paula_SetDMACON
	
	; now set new data/length for after first Paula cycle
		
	; set Paula data
	mov eax,n_loopstart[edi]
	call _Paula_SetDat

	; set Paula length
	movzx eax,word ptr n_replen[edi]
	jmp _Paula_SetLen

VolumeFineUp:
	cmp byte ptr m_tick[ebp],0
	jne Return
	and al,15
	jmp VolSlideUp

VolumeFineDown:
	cmp byte ptr m_tick[ebp],0
	jne Return
	and al,15
	jmp VolSlideDown2

NoteCut:
	cmp al,m_tick[ebp]
	jne Return
	mov byte ptr n_volume[edi],0
	retn

NoteDelay:
	cmp al,m_tick[ebp]
	jne Return
	cmp byte ptr n_note[edi],0
	je Return
	; -------------------------------
	jmp DoRetrg

PatternDelay:
	cmp byte ptr m_tick[ebp],0
	jne Return
	cmp byte ptr PattDelayTime2,0
	jne Return
	inc al
	mov PattDelayTime,al
	retn

FunkIt:
	cmp byte ptr m_tick[ebp],0
	jne Return
	shl al,4
	and byte ptr n_glissfunk[edi],15
	or byte ptr n_glissfunk[edi],al
	jz Return
UpdateFunk:
	movzx eax,byte ptr n_glissfunk[edi]
	shr eax,4
	jz Return
	mov al,[FunkTable+eax]
	add n_funkoffset[edi],al
	test byte ptr n_funkoffset[edi],128
	jz Return
	mov byte ptr n_funkoffset[edi],0
	; --------------------------------
	; non-PT sanity check (test pointers)
	mov eax,n_loopstart[edi]
	test eax,eax
	jz Return
	mov ecx,n_wavestart[edi]
	test ecx,ecx
	jz Return
	; --------------------------------
	movzx ebx,word ptr n_replen[edi]
	add eax,ebx
	add eax,ebx
	inc ecx
	cmp ecx,eax
	jb funkok
	mov ecx,n_loopstart[edi]
funkok:	mov n_wavestart[edi],ecx
	not byte ptr [ecx]	; [ecx] = -1 - [ecx]
	retn

_Replayer_Tick ENDP

.DATA

;=============================================================================
;                                 JUMP TABLES
;=============================================================================
	ALIGN 4

JumpList1: ; tick > 0
	dd Arpeggio		; 0xy (Arpeggio)
	dd PortaUp		; 1xx (Portamento Up)
	dd PortaDown		; 2xx (Portamento Down)
	dd TonePortamento	; 3xx (Tone Portamento)
	dd Vibrato		; 4xy (Vibrato)
	dd TonePlusVolSlide	; 5xy (Tone Portamento + Volume Slide)
	dd VibratoPlusVolSlide	; 6xy (Vibrato + Volume Slide)
	dd SetBack		; 7 - not used here
	dd SetBack		; 8 - unused!
	dd SetBack		; 9 - not used here
	dd SetBack		; A - not used here
	dd SetBack		; B - not used here
	dd SetBack		; C - not used here
	dd SetBack		; D - not used here
	dd E_Commands		; Exy (Extended Commands)
	dd SetBack		; F - not used here

JumpList2: ; tick = 0
	dd PerNop		; 0 - not used
	dd PerNop		; 1 - not used
	dd PerNop		; 2 - not used
	dd PerNop		; 3 - not used
	dd PerNop		; 4 - not used
	dd PerNop		; 5 - not used
	dd PerNop		; 6 - not used
	dd PerNop		; 7 - not used
	dd PerNop		; 8 - not used
	dd SampleOffset		; 9xx (Set Sample Offset)
	dd PerNop		; A - not used
	dd PositionJump		; Bxx (Position Jump)
	dd VolumeChange		; Cxx (Set Volume)
	dd PatternBreak		; Dxx (Pattern Break)
	dd E_Commands		; Exy (Extended Commands)
	dd SetSpeed		; Fxx (Set Speed)

E_JumpList: ; tick = 0
	dd FilterOnOff		; E0x (Set LED Filter)
	dd FinePortaUp		; E1x (Fine Portamento Up)
	dd FinePortaDown	; E2x (Fine Portamento Down)
	dd SetGlissControl	; E3x (Glissando/Funk Control)
	dd SetVibratoControl	; E4x (Vibrato Control)
	dd SetFineTune		; E5x (Set Finetune)
	dd JumpLoop		; E6x (Pattern Loop)
	dd SetTremoloControl	; E7x (Tremolo Control)
	dd KarplusStrong	; E8x (Karplus-Strong)
	dd RetrigNote		; E9x (Retrig Note)
	dd VolumeFineUp		; EAx (Fine Volume-Slide Up)
	dd VolumeFineDown	; EBx (Fine Volume-Slide Down)
	dd NoteCut		; ECx (Note Cut)
	dd NoteDelay		; EDx (Note Delay)
	dd PatternDelay		; EEx (Pattern Delay)
	dd FunkIt		; EFx (Invert Loop)

;=============================================================================
;                                   TABLES
;=============================================================================
	ALIGN 4
; Used for increasing the ticks in the playback counter (based on current BPM).
;
; Formula:
; for (int32_t i = 32; i <= 255; i++)
; {
;     const double dHz = ciaBpm2Hz(bpm);
;     musicTimeTab32[i-32] = (uint32_t)round((UINT32_MAX+1.0) / dHz);
; }
;
; // vblank mode (~49.92Hz)
; musicTimeTab32[256] = (uint32_t)round((UINT32_MAX+1.0) / AMIGA_PAL_VBLANK_HZ);
;
MusicTimeTab32:
        dd 1400121Ch,1364DD27h,12D2E542h,12492643h,11C729EAh,114C3302h
        dd 10D79BFEh,10691DE9h,10001337h,0F9C1D4Dh,0F3CF538h,0EE23C5Ch
        dd 0E8BABC7h,0E38E4DFh,0DE9CFFCh,0D9DF6DEh,0D555985h,0D0FB0FEh
        dd 0CCCCDFCh,0C8C98D7h,0C4ECA9Ch,0C13634Bh,0BDA1BF1h,0BA2F48Dh
        dd 0B6DBDD3h,0B3A77C3h,0B08DB68h,0AD90069h,0AAAB77Ah,0A7DE8F4h
        dd 0A5294D6h,0A28A37Ah,0A0014E1h,09D8A216h,09B27A67h,098D6E86h
        dd 09697E74h,0946928Ah,09249321h,0903AF87h,08E3A0C8h,08C466E4h
        dd 08A61981h,08888952h,086BCDFFh,084FCFE0h,08348EF5h,0819F398h
        dd 0800156Fh,07E6C52Eh,07CE1A7Ah,07B5FDAEh,079E866Fh,07879D18h
        dd 07712A01h,075B44D2h,0745D5E4h,0730F4DCh,071C726Fh,07087DEAh
        dd 06F4E7FEh,06E1C853h,06CF0742h,06BCA4CCh,06AAB896h,06991354h
        dd 0687E452h,0676FC45h,066672D1h,06563051h,06464C6Bh,0636AF79h
        dd 06277121h,06186216h,0609B1A6h,05FB4828h,05ED0DF9h,05DF3263h
        dd 05D1861Ah,05C420C5h,05B6EABDh,05A9FBA9h,059D3BE1h,0590C30Eh
        dd 05847987h,05785F4Eh,056C8C08h,0560D069h,05555BBDh,054A165Fh
        dd 053F004Dh,05341989h,05294A6Bh,051EC241h,051451BDh,050A282Dh
        dd 05001644h,04F61C01h,04EC510Bh,04E2B562h,04D94907h,04CFF452h
        dd 04C6B743h,04BDA982h,04B4CB0Dh,04ABEC99h,04A35518h,049ABD98h
        dd 04925564h,048A04D7h,0481E397h,0479C257h,0471D064h,0469F618h
        dd 04623372h,045AA019h,04530CC0h,044BA8B5h,044444A9h,043D0FEBh
        dd 0435F2D3h,042EED61h,0427E7F0h,042111CBh,041A534Eh,041394D0h
        dd 040D059Fh,0406766Eh,0400168Bh,03F9B6A7h,03F36E6Ah,03ED3DD3h
        dd 03E70D3Dh,03E10BF4h,03DB0AAAh,03D52107h,03CF4F0Bh,03C97D0Eh
        dd 03C3DA5Fh,03BE37B0h,03B89501h,03B3219Eh,03ADAE3Ch,03A83ADAh
        dd 03A2F6C5h,039DB2B0h,03988641h,039359D3h,038E450Bh,038947E9h
        dd 03844AC8h,037F4DA7h,037A7FD2h,0375B1FEh,0370E42Ah,036C2DFCh
        dd 03678F74h,0362F0EDh,035E5266h,0359E32Ch,03555C4Bh,0350ED11h
        dd 034C957Dh,03483DEAh,0343FDFCh,033FBE0Fh,033B7E22h,033755DCh
        dd 0333453Ch,032F349Ch,032B23FCh,03272B02h,03233209h,031F390Fh
        dd 031B57BDh,03178E10h,0313C464h,030FFAB7h,030C310Bh,03087F05h
        dd 0304E4A6h,03014A47h,02FDAFE8h,02FA1588h,02F692D0h,02F31017h
        dd 02EFA505h,02EC224Ch,02E8CEE0h,02E563CEh,02E21062h,02DEBCF7h
        dd 02DB8132h,02D82DC6h,02D509A8h,02D1CDE2h,02CEA9C4h,02CB85A5h
        dd 02C86187h,02C5550Fh,02C24897h,02BF3C1Fh,02BC2FA7h,02B93AD5h
        dd 02B64604h,02B35133h,02B07408h,02AD96DDh,02AAB9B2h,02A7DC87h
        dd 02A51703h,02A2517Eh,029F8BFAh,029CC675h,029A1898h,02975313h
        dd 0294A535h,02920EFEh,028F6120h,028CCAE9h,028A34B2h,02879E7Bh
        dd 02851FEAh,028289B3h 
        dd 0520CF4Ch ; ~49.92Hz (vblank tempo mode)

FunkTable:
	db 0,5,6,7,8,10,11,13,16,19,22,26,32,43,64,128

VibratoTable:
	db   0, 24, 49, 74, 97,120,141,161
	db 180,197,212,224,235,244,250,253
	db 255,253,250,244,235,224,212,197
	db 180,161,141,120, 97, 74, 49, 24

	ALIGN 2
PeriodTable:
ftune0:
	dw 856,808,762,720,678,640,604,570,538,508,480,453
	dw 428,404,381,360,339,320,302,285,269,254,240,226
	dw 214,202,190,180,170,160,151,143,135,127,120,113,0
ftune1:
	dw 850,802,757,715,674,637,601,567,535,505,477,450
	dw 425,401,379,357,337,318,300,284,268,253,239,225
	dw 213,201,189,179,169,159,150,142,134,126,119,113,0
ftune2:
	dw 844,796,752,709,670,632,597,563,532,502,474,447
	dw 422,398,376,355,335,316,298,282,266,251,237,224
	dw 211,199,188,177,167,158,149,141,133,125,118,112,0
ftune3:
	dw 838,791,746,704,665,628,592,559,528,498,470,444
	dw 419,395,373,352,332,314,296,280,264,249,235,222
	dw 209,198,187,176,166,157,148,140,132,125,118,111,0
ftune4:
	dw 832,785,741,699,660,623,588,555,524,495,467,441
	dw 416,392,370,350,330,312,294,278,262,247,233,220
	dw 208,196,185,175,165,156,147,139,131,124,117,110,0
ftune5:
	dw 826,779,736,694,655,619,584,551,520,491,463,437
	dw 413,390,368,347,328,309,292,276,260,245,232,219
	dw 206,195,184,174,164,155,146,138,130,123,116,109,0
ftune6:
	dw 820,774,730,689,651,614,580,547,516,487,460,434
	dw 410,387,365,345,325,307,290,274,258,244,230,217
	dw 205,193,183,172,163,154,145,137,129,122,115,109,0
ftune7:
	dw 814,768,725,684,646,610,575,543,513,484,457,431
	dw 407,384,363,342,323,305,288,272,256,242,228,216
	dw 204,192,181,171,161,152,144,136,128,121,114,108,0
ftune8:
	dw 907,856,808,762,720,678,640,604,570,538,508,480
	dw 453,428,404,381,360,339,320,302,285,269,254,240
	dw 226,214,202,190,180,170,160,151,143,135,127,120,0
ftune9:
	dw 900,850,802,757,715,675,636,601,567,535,505,477
	dw 450,425,401,379,357,337,318,300,284,268,253,238
	dw 225,212,200,189,179,169,159,150,142,134,126,119,0
ftuneA:
	dw 894,844,796,752,709,670,632,597,563,532,502,474
	dw 447,422,398,376,355,335,316,298,282,266,251,237
	dw 223,211,199,188,177,167,158,149,141,133,125,118,0
ftuneB:
	dw 887,838,791,746,704,665,628,592,559,528,498,470
	dw 444,419,395,373,352,332,314,296,280,264,249,235
	dw 222,209,198,187,176,166,157,148,140,132,125,118,0
ftuneC:
	dw 881,832,785,741,699,660,623,588,555,524,494,467
	dw 441,416,392,370,350,330,312,294,278,262,247,233
	dw 220,208,196,185,175,165,156,147,139,131,123,117,0
ftuneD:
	dw 875,826,779,736,694,655,619,584,551,520,491,463
	dw 437,413,390,368,347,328,309,292,276,260,245,232
	dw 219,206,195,184,174,164,155,146,138,130,123,116,0
ftuneE:
	dw 868,820,774,730,689,651,614,580,547,516,487,460
	dw 434,410,387,365,345,325,307,290,274,258,244,230
	dw 217,205,193,183,172,163,154,145,137,129,122,115,0
ftuneF:
	dw 862,814,768,725,684,646,610,575,543,513,484,457
	dw 431,407,384,363,342,323,305,288,272,256,242,228
	dw 216,203,192,181,171,161,152,144,136,128,121,114,0

	; Arpeggio on -1 finetuned samples can do an overflown read from
	; the period table. Here's the correct overflow values from the
	; "CursorPosTable" and "UnshiftedKeymap" table, which are located
	; right after the period table. These tables and their order didn't
	; seem to change in the different PT1.x/PT2.x versions (I checked
	; the source codes).
	; PS: This is not a guess, these values *are* correct!
	dw 774,1800,2314,3087,4113,4627,5400,6426,6940,7713
	dw 8739,9253,24625,12851,13365

	ALIGN 4
TunedPeriodPtr:
	dd ftune0,ftune1,ftune2,ftune3,ftune4,ftune5,ftune6,ftune7
	dd ftune8,ftune9,ftuneA,ftuneB,ftuneC,ftuneD,ftuneE,ftuneF

;=============================================================================
;                                 VARIABLES
;=============================================================================
	ALIGN 4

PattReadPtr:		dd 0
PatternPtr:		dd 0
DMACONtemp:		dw 0
ActiveChannels:		dw 15
PBreakFlag:		db 0
PosJumpAssert:		db 0
SongPosition:		db 0
PBreakPosition:		db 0
PattDelayTime:		db 0
PattDelayTime2:		db 0
PatternPosition:	db 0
SetBPMFlag:		db 0
StepPlayEnable:		db 0

END
