#ifndef __REPLAYER_H__
#define __REPLAYER_H__

#include <stdint.h>
#include <stdbool.h>

#define PT_MAX_PATTERNS 100

#define PT_MIN_BPM 32
#define PT_MAX_BPM 255

enum
{
	PLAY_MODE_IDLE = 0,
	PLAY_MODE_SONG = 1,
	PLAY_MODE_PATTERN = 2,
	PLAY_MODE_EDIT = 3,
	PLAY_MODE_RECORD = 4
};

// loader error codes
enum
{
	LERR_SUCCESS    = 0,
	LERR_FILE       = 1,
	LERR_IO         = 2,
	LERR_NOT_A_MOD  = 3,
	LERR_OUT_OF_MEM = 4
};

typedef struct
{
	uint8_t note, smp, efx, efxData;
} note_t;

typedef struct
{
	// accessed by areplay.asm

	// BYTE
	bool useVblankTempo;
	uint8_t numPatterns; // WORD
	uint8_t length, row, order, pattern, tick, speed, BPM;
	uint8_t muted[4];
	uint8_t orderList[128];

	// accessed by C code
	uint32_t modSize;
	char name[20+1];
} ptMod_t;

typedef struct
{
	// accessed by areplay.asm
	int8_t *data; // DWORD
	uint16_t length, loopStart, loopLength; // WORD
	int8_t finetune; // BYTE
	uint8_t volume; // BYTE

	// accessed by C code
	char name[22+1];
} ptSmp_t;

typedef struct
{
	uint8_t n_note, n_samplenum, n_cmd, n_param;
	int8_t *n_start, *n_wavestart, *n_loopstart;
	uint16_t n_length, n_replen;
	int16_t n_period, n_wantedperiod;
	uint32_t n_chanindex;
	int8_t n_volume;
	uint8_t n_toneportdirec, n_pattpos, n_loopcount;
	uint8_t n_wavecontrol, n_glissfunk, n_sampleoffset, n_toneportspeed;
	uint8_t n_vibratocmd, n_tremolocmd, n_finetune, n_funkoffset;
	uint8_t n_vibratopos, n_tremolopos, n_muted;
	uint16_t n_dmabit;
} ptChn_t;

extern volatile bool Replayer_ModuleLoaded, Replayer_SongPlaying;
extern uint8_t Replayer_PlayMode;
extern ptMod_t module;
extern ptSmp_t samples[31];
extern ptChn_t channel[4];
extern note_t *patternData;
extern uint32_t Replayer_PlaybackSeconds, Replayer_PlaybackSecondsFrac;

void Replayer_Update(uint32_t row, uint32_t order);
void Replayer_CalculateModSize(void);
void Replayer_SetBPM(uint32_t bpm);
bool Replayer_Init(void);
void Replayer_Close(void);
bool Replayer_InitBPMTable(void);
void Replayer_FreeBPMTable(void);
void Replayer_SetPos(uint8_t pos);
void Replayer_Free(void);
bool Replayer_Load(const char *filename);
uint8_t Replayer_GetLoadError(void);
void Replayer_Play(uint8_t playMode, uint8_t row);
void Replayer_Stop(void);

// areplay.asm

extern void Replayer_ConvertPatternData(note_t *dataPtr, uint32_t numPatterns);
#pragma aux Replayer_ConvertPatternData parm [ESI] [EAX];

extern void Replayer_PrepareSetPos(uint32_t order);
#pragma aux Replayer_PrepareSetPos parm [EAX];

extern void Replayer_StepPlayForwards(void);
extern void Replayer_StepPlayBackwards(void);
extern void Replayer_PreparePlay(void);
extern void Replayer_PrepareStop(void);
extern void Replayer_Tick(void);

#endif
