.386p
.MODEL FLAT

	; Interrupt routines for PIT timer chip.
	; Some of this code is partly based on Fasttracker II's code
	;
	; wasm's lack of local labels is disgusting! Hence why the
	; labels are so bad.
	;

AUDIO_BUF_SIZE EQU 8192 ; must match the value in audio.h!

	; offsets into audio struct (audio.h)
a_soundReady		EQU 0
a_dmaBuffer		EQU 4
a_sampleShift		EQU 8
a_dmaNum		EQU 9
a_dmaCountPort		EQU 10
a_dmaFlipflopPort	EQU 11

	EXTERN __GETDS:PROC

	; main.c
	EXTERN _audioStructPtr:DWORD
	
	; audio.c
	EXTERN _Audio_MixSamples:PROC
	EXTERN _Audio_MixSamplesDummy:PROC

	; ui.c
	EXTERN _showRasterbars:BYTE	

	; pit.c
	EXTERN _dmaMixPos:DWORD
	EXTERN _oldInt8VectorDelta:DWORD
	EXTERN _oldInt8VectorFrac:DWORD
	EXTERN _samplesPerFrame15:DWORD
	EXTERN _samplesPerFrame25:DWORD	
	EXTERN _oldInt8Vector:PROC

	; video.c
	EXTERN _framesPassed:DWORD
	EXTERN _Video_WaitTimeout:PROC

	; dma.c
	EXTERN _dma:DWORD
	EXTERN _dmaPorts:DWORD

	; avideo.asm
	EXTERN vertIntTimeOut:DWORD
	EXTERN oneFrameTime1:WORD
	EXTERN oneFrameTime2:WORD

	; mouse.c
	EXTERN _updateMousePalette:BYTE
	EXTERN _mouseR:BYTE
	EXTERN _mouseG:BYTE
	EXTERN _mouseB:BYTE

	PUBLIC _int8Vector
	PUBLIC _int8VectorNoAudio

.CODE

SET_BG_COLOR MACRO r,g,b
	mov al,0FFh
	mov dx,3C6h
	out dx,al
	add edx,2	
	xor eax,eax
	out dx,al
	inc edx
	mov al,r
	out dx,al
	mov al,g
	out dx,al
	mov al,b
	out dx,al
ENDM

RESTORE_BG_COLOR MACRO
	mov al,0FFh
	mov dx,3C6h
	out dx,al
	add edx,2
	xor eax,eax
	out dx,al
	inc edx
	out dx,al
	out dx,al
	out dx,al
ENDM

	ALIGN 4
_int8Vector PROC FAR
	pushad
	push ds
	push es
	call __GETDS	
	mov ax,ds
	mov es,ax
	; --------------------------------------------
	; If inactive,
	;   wait until end, use: oneFrameTime1
	; else
	;   use: oneFrameTime2
	; --------------------------------------------
	mov dx,3DAh
	; --------------------------------------------
	mov ecx,vertIntTimeOut
L00:	in al,dx
	test al,1
	loopnz L00
	jz BelowVBL
	; --------------------------------------------
StillDoingVBL:
	mov ecx,vertIntTimeOut
L01:	in al,dx
	test al,1
	loopnz L01
	jnz StillDoingVBL
	mov cx,oneFrameTime1	; oneFrameTime - (oneFrameTime/128)
	jmp SetPITCount
	; --------------------------------------------
BelowVBL:
	mov cx,oneFrameTime2	; oneFrameTime - (oneFrameTime/16)
	; --------------------------------------------
SetPITCount:
	mov al,36h
	out 43h,al
	mov al,cl
	out 40h,al
	mov al,ch
	out 40h,al
	; --------------------------------------------
	xor eax,eax	; al = 0
	xchg al,_updateMousePalette
	test al,al
	jz L03
	; --------------------------------------------
	; Update mouse palette
	; --------------------------------------------
	mov dx,3C8h
	mov al,16
	out dx,al
	inc edx
	mov ecx,8
L02:	mov al,_mouseR
	out dx,al
	mov al,_mouseG
	out dx,al
	mov al,_mouseB
	out dx,al
	dec ecx
	jnz L02
	; --------------------------------------------
L03:	inc dword ptr _framesPassed
	; --------------------------------------------
	cmp byte ptr _showRasterbars,0
	je L04
	SET_BG_COLOR 63,0,0
L04:	; --------------------------------------------
	; run old interrupt 8 at old PIT channel 0 rate (18.2Hz)
	mov eax,_oldInt8VectorDelta
	add _oldInt8VectorFrac,eax
	jnc L05
	pushfd
	call fword ptr _oldInt8Vector
	jmp L06
L05:	; send EOI to PIC chip #1 (IRQ 0)
	mov al,20h
	out 20h,al
L06:	; --------------------------------------------
	mov esi,_audioStructPtr
	cmp byte ptr a_soundReady[esi],1
	jne NoMix
	; --------------------------------------------
	; Read sound DMA position.
	;
	; Read it until the difference between two reads
	; is < 32, this is because of memory managers etc.
	; --------------------------------------------
	movzx ebx,byte ptr a_dmaCountPort[esi]
	movzx ecx,byte ptr a_dmaFlipflopPort[esi]
	; ------------
dmaPosLoop:
	mov edx,ecx
	xor eax,eax
	out dx,al	; clear flip-flop for 16-bit read
	; ------------
	mov edx,ebx
	in al,dx
	mov ah,al
	in al,dx
	xchg al,ah
	mov edi,eax
	; ------------
	mov edx,ecx
	xor eax,eax
	out dx,al	; clear flip-flop for 16-bit read
	; ------------
	mov edx,ebx
	in al,dx
	mov ah,al
	in al,dx
	xchg al,ah
	; ------------
	sub edi,eax
	jns notSigned
	neg edi
notSigned:
	; ------------
	cmp edi,32	; difference < 32?
	jnb dmaPosLoop	; nope, go again
	; ------------
	cmp byte ptr a_dmaNum[esi],4
	jb no16Bit
	add eax,eax
no16Bit:
	neg eax
	mov ebp,AUDIO_BUF_SIZE-1
	add eax,ebp
	mov ebx,eax
	; --------------------------------------------
	; EBX = audio DMA position
	; --------------------------------------------
	mov al,1
	xchg al,mixingOngoing
	cmp al,1
	je L07
	; --------------------------------------------
	mov edx,_dmaMixPos
	mov eax,edx		; eax = copy of _dmaMixPos
	sub edx,ebx
	and edx,ebp
	mov cl,a_sampleShift[esi]
	shr edx,cl
	mov ebx,_samplesPerFrame25
	sub ebx,edx		; ebx = samplesToMix
	jb NoMix
	; --------------------------------------------
	cmp ebx,_samplesPerFrame15
	ja NoMix
	; --------------------------------------------
	and ebx,0-16		; align to blocks of 16
	jz NoMix
	; --------------------------------------------
	mov edi,a_dmaBuffer[esi]
	mov edx,ebx
	shl edx,cl
	add edx,eax
	and edx,ebp
	mov _dmaMixPos,edx
	call _Audio_MixSamples
NoMix:	; --------------------------------------------
	mov byte ptr mixingOngoing,0
	; --------------------------------------------
L07:	cmp byte ptr _showRasterbars,0
	je L08
	RESTORE_BG_COLOR
L08:	; --------------------------------------------
	pop es
	pop ds
	popad
	iretd
_int8Vector ENDP

	ALIGN 4
_int8VectorNoAudio PROC FAR
	pushad
	push ds
	push es
	call __GETDS
	mov ax,ds
	mov es,ax
	; --------------------------------------------
	; If inactive,
	;   wait until end, use: oneFrameTime1
	; else
	;   use: oneFrameTime2
	; --------------------------------------------
	mov dx,3DAh
	; --------------------------------------------
	mov ecx,vertIntTimeOut
L09:	in al,dx
	test al,1
	loopnz L09
	jz BelowVBL2
	; --------------------------------------------
StillDoingVBL2:
	mov ecx,vertIntTimeOut
L10:	in al,dx
	test al,1
	loopnz L10
	jnz StillDoingVBL2
	mov cx,oneFrameTime1	; oneFrameTime - (oneFrameTime/128)
	jmp SetPITCount2
	; --------------------------------------------
BelowVBL2:
	mov cx,oneFrameTime2	; oneFrameTime - (oneFrameTime/16)
	; --------------------------------------------
SetPITCount2:
	mov al,36h
	out 43h,al
	mov al,cl
	out 40h,al
	mov al,ch
	out 40h,al
	; --------------------------------------------	
	xor eax,eax	; al = 0
	xchg al,_updateMousePalette
	test al,al
	jz L12
	; --------------------------------------------
	; Update mouse palette
	; --------------------------------------------
	mov dx,3C8h
	mov al,16
	out dx,al
	inc edx
	mov ecx,8
L11:	mov al,_mouseR
	out dx,al
	mov al,_mouseG
	out dx,al
	mov al,_mouseB
	out dx,al
	dec ecx
	jnz L11
	; --------------------------------------------
L12:	inc dword ptr _framesPassed
	; --------------------------------------------
	cmp byte ptr _showRasterbars,0
	je L13
	SET_BG_COLOR 63,0,0
L13:	; --------------------------------------------
	; run old interrupt 8 at old PIT channel 0 rate (18.2Hz)
	mov eax,_oldInt8VectorDelta
	add _oldInt8VectorFrac,eax
	jnc L14
	pushfd
	call fword ptr _oldInt8Vector
	jmp L15
L14:	; send EOI to PIC chip #1 (IRQ 0)
	mov al,20h
	out 20h,al
L15:	; --------------------------------------------
	mov eax,512 ; 2^9 (extra precision)
	call _Audio_MixSamplesDummy
	; --------------------------------------------
	cmp byte ptr _showRasterbars,0
	je L16
	RESTORE_BG_COLOR
L16:	; --------------------------------------------
	pop es
	pop ds
	popad
	iretd
_int8VectorNoAudio ENDP

mixingOngoing:		db 0

END
