#include <stdio.h>
#include <stdint.h>
#include <stdbool.h>
#include <dos.h>
#include <conio.h>
#include <i86.h>
#include "../ui.h"
#include "../audio/replayer.h"
#include "../gfx/font.h"
#include "video.h"
#include "mouse.h"

#define MOUSE_DRIVER_INSTALLED 0xFFFF

static bool mouseOpen;
static uint8_t blockCounter;

mouse_t mouse;

// amouse.asm
extern void __interrupt Mouse_Interrupt(void);

void Mouse_SetDiskAccessColor(void)
{
	Video_SetMouseColor(21, 55, 8);
}

void Mouse_UpdateColor(void)
{
	if (mouse.blockInput)
		Video_SetMouseColor(55, 0, 4);
	else if (ui.askBoxShown)
		Video_SetMouseColor(51, 8, 38);
	else if (Replayer_PlayMode == PLAY_MODE_SONG || Replayer_PlayMode == PLAY_MODE_PATTERN)
		Video_SetMouseColor(63, 29, 4);
	else if (Replayer_PlayMode == PLAY_MODE_RECORD || Replayer_PlayMode == PLAY_MODE_EDIT)
		Video_SetMouseColor(0, 46, 63);
	else // idle
		Video_SetMouseColor(63, 63, 63);
}

void Mouse_Error(void)
{
	mouse.blockInput = true;
	Mouse_UpdateColor();
	blockCounter = ERR_DURATION;
}

bool Mouse_Initialize(void)
{
	struct SREGS sr;
	union REGS r;
	
	// check if interrupt 0x33 is installed before interacting with it
	segread(&sr);
	r.w.ax = 0x3533;
	int386x(0x21, &r, &r, &sr);
	if (sr.es != 0)
	{
		// detect (and reset) mouse
		r.w.ax = 0x0000;
		int386(0x33, &r, &r);

		if (r.w.ax == MOUSE_DRIVER_INSTALLED)
		{
			// mouse is present, center internal mouse coordinates
			mouse.x = (SCREEN_W/2) - (POINTER_W/2);
			mouse.y = (SCREEN_H/2) - (POINTER_H/2);

			// set mouse callback
			segread(&sr);
			r.w.ax = 0x000C;
			r.w.cx = 0x001F; // event mask (pos change, buttons pressed/released)
			r.x.edx = FP_OFF(Mouse_Interrupt);
			sr.es = FP_SEG(Mouse_Interrupt);
			int386x(0x33, &r, &r, &sr);

			mouseOpen = true;
			return true;
		}
	}

	printf("ERROR: Couldn't detect an installed mouse driver!\n");
	return false;
}

void Mouse_Close(void)
{
	struct SREGS sr;
	union REGS r;

	if (mouseOpen)
	{
		// disable mouse callback
		segread(&sr);
		r.w.ax = 0x000C;
		r.w.cx = 0;
		r.x.edx = FP_OFF(Mouse_Interrupt);
		sr.es = FP_SEG(Mouse_Interrupt);
		int386x(0x33, &r, &r, &sr);

		// reset mouse
		r.w.ax = 0x0000;
		int386(0x33, &r, &r);

		mouseOpen = false;
	}
}

void Mouse_HandleInput(void)
{
	// handle busy mouse (red color for 1 second + input blocking)
	if (mouse.blockInput)
	{
		blockCounter--;
		if (blockCounter == 0)
		{	
			mouse.blockInput = false;
			Mouse_UpdateColor();
		}

		return;
	}

	if (!mouse.clickHandled && mouse.leftButtonDown)
	{
		mouse.clickHandled = true;
		UI_HandleMouseClick();
	}
}
