#include <stdio.h>
#include <stdint.h>
#include <stdbool.h>
#include <string.h>
#include <stdlib.h>
#include <conio.h>
#include "audio/replayer.h"
#include "hardware/video.h"
#include "gfx/font.h"
#include "gfx/drawcode.h"
#include "ui.h"
#include "pattern.h"

#define TOP_ROWS 6
#define BOTTOM_ROWS 6
#define PAT_CURSOR_Y 192

static bool cursorWritten = false;

static const uint8_t cursorXPos[6*4] =
{
	 4,  7,  8,  9, 10, 11, // channel #1
	13, 16, 17, 18, 19, 20, // channel #2
	22, 25, 26, 27, 28, 29, // channel #3
	31, 34, 35, 36, 37, 38  // channel #4
};

void Pattern_ClearCache(void)
{
	__asm
	{
		push eax
		push ecx
		push edi
		mov eax,0xFFFFFFFF
		mov ecx,VISIBLE_ROWS*4*4
		mov edi,oldRows
		rep stosd
		pop edi
		pop ecx
		pop eax
	};
}

void Pattern_RedrawCursor(void)
{
	cursorWritten = false;
	Pattern_SetCursorPos(ui.patt.cursorPos);
}

void Pattern_SetCursorPos(uint8_t cursorPos)
{
	uint8_t *ptr1, *ptr2, i;

	if (cursorPos > 23)
		cursorPos = 23;

	if (!ui.samplerScreenShown) // draw cursor
	{
		// erase old cursor first
		if (cursorPos != ui.patt.cursorPos && cursorWritten)
		{
			ptr1 = (uint8_t *)VGA_FRAMEBUFFER + (cursorXPos[ui.patt.cursorPos]+(PAT_CURSOR_Y*(SCREEN_W/8)));
			ptr2 = ptr1 + 11*(SCREEN_W/8);

			if ((ui.patt.cursorPos % 6) == 0) // was on a note?
			{
				Video_SetPlaneBitmask(5);
				ptr1[0] = 0x00;
				ptr1[1] = 0x00;
				ptr1[2] = 0x00;
				ptr1[0+(SCREEN_W/8)] = 0x00;
				ptr1[1+(SCREEN_W/8)] = 0x00;
				ptr1[2+(SCREEN_W/8)] = 0x00;

				Video_SetPlaneBitmask(3);
				ptr1[0] = 0xFF;
				ptr1[1] = 0xFF;
				ptr1[2] = 0xFF;
				ptr1[0+(SCREEN_W/8)] = 0xFF;
				ptr1[1+(SCREEN_W/8)] = 0xFF;
				ptr1[2+(SCREEN_W/8)] = 0xFF;
			}
			else
			{
				Video_SetPlaneBitmask(5);
				ptr1[0] = 0x00;
				ptr1[0+(SCREEN_W/8)] = 0x00;

				Video_SetPlaneBitmask(3);
				ptr1[0] = 0xFF;
				ptr1[0+(SCREEN_W/8)] = 0xFF;
			}
		}

		ptr1 = (uint8_t *)VGA_FRAMEBUFFER + (cursorXPos[cursorPos]+(PAT_CURSOR_Y*(SCREEN_W/8)));
		ptr2 = ptr1 + 11*(SCREEN_W/8);

		if ((cursorPos % 6) == 0) // on a note?
		{
			Video_SetPlaneBitmask(3);
			ptr1[0] = 0x00;
			ptr1[1] = 0x00;
			ptr1[2] = 0x00;
			ptr1[0+(SCREEN_W/8)] = 0x00;
			ptr1[1+(SCREEN_W/8)] = 0x00;
			ptr1[2+(SCREEN_W/8)] = 0x00;
			
			Video_SetPlaneBitmask(5);
			ptr1[0] = 0xFF;
			ptr1[1] = 0xFF;
			ptr1[2] = 0xFF;
			ptr1[0+(SCREEN_W/8)] = 0xFF;
			ptr1[1+(SCREEN_W/8)] = 0xFF;
			ptr1[2+(SCREEN_W/8)] = 0xFF;
		}
		else
		{
			Video_SetPlaneBitmask(3);
			ptr1[0] = 0x00;
			ptr1[0+(SCREEN_W/8)] = 0x00;

			Video_SetPlaneBitmask(5);
			ptr1[0] = 0xFF;
			ptr1[0+(SCREEN_W/8)] = 0xFF;
		}

		cursorWritten = true;
	}

	ui.patt.cursorPos = cursorPos;
}

void Pattern_Draw(int32_t pattern)
{
	uint8_t *vramPtr;
	int32_t row, currRow, rows, endRow, startRow;

	currRow = module.row;
	startRow = 0;

	row = currRow - TOP_ROWS;
	if (row < 0)
		row = 0;

	endRow = row + VISIBLE_ROWS;

	Video_SetPlaneBitmask(7);

	// clear top/bottom outside rows
	if (currRow < TOP_ROWS)
	{
		startRow = TOP_ROWS - currRow;
		endRow -= startRow;

		Pattern_ClearRows(0, startRow);
	}
	else if (currRow > 63-BOTTOM_ROWS)
	{
		rows = currRow-(63-BOTTOM_ROWS);
		endRow -= rows;	

		Pattern_ClearRows(TOP_ROWS+(64-currRow), rows);
	}

	Pattern_DrawRows(endRow, row, startRow, &patternData[(pattern << 8) + (row << 2)]);
}
